#code for Chapter 4 "Propensity Score Stratification" of book:
#Leite, W. L. (2016). Practical propensity score analysis using R. 
#Thousand Oaks, CA: Sage. 
# 
#this is the code that was used to generate the example results in the book
#As the R software and the R packages used in this example are updated frequently
#some incompatibilities between the current code and new R versions or package versions
#may appear
#Any updates to the code will be posted at:
# http://www.practicalpropensityscore.com


#=========================================================
#this code is for estimating the effect of schools 
#having at least one full-time security guards
#on the frequency of harsh punishment
#using propensity score stratification
#using data from the School Survey on Crime and Safety

#Read dataset
load(file="Chapter4_ssocs08.Rdata")


#Get predictors of treatment assignment
#variable C0116 (metal detectors) was excluded because 97.4% of the schools 
#do not have students pass through metal detectors and only 1 of the untreated
#school have them pass through metal detectors
covariateNames <- names(SSOCS.data)[c(5:6,15:27,29,31,33,35:37,39:54,80:85,150:155,162:164,178,
                                      197:200, 417:422)]

#convert missing value indicators into binary variables
for (var in 252:414) {
SSOCS.data[,var] <- ifelse(SSOCS.data[,var] == 0, 0, 1) }

#obtain proportions of missing data for variables of interest
missingIndicatorNames <- paste("I", covariateNames, sep="")
missingIndicatorNames <- missingIndicatorNames[-c(53:57)] #remove composite variables
propMissing = apply(SSOCS.data[,missingIndicatorNames],2,mean)
missingIndicatorNames <- missingIndicatorNames[propMissing > 0.05]

#obtain imputation flags for variables of interest
#check whether any dummy missing indicators are redundant because
#variables have missing values for the same cases
missingCorrelations <- cor(SSOCS.data[,missingIndicatorNames])
diag(missingCorrelations) <- 0
maxCorrelations <- apply(missingCorrelations,2,max) 
dummyNAnames <- names(maxCorrelations)[maxCorrelations < 0.8]
maxCorrelationsHigh <- maxCorrelations[!duplicated(maxCorrelations)]
dummyNAnames <- c(dummyNAnames,names(maxCorrelationsHigh)[maxCorrelationsHigh>=0.8])



#create propensity score estimation formula
covariateNames <- c(covariateNames,dummyNAnames)
psFormula <- paste(covariateNames, collapse="+")
psFormula <- formula(paste("treat~",psFormula, sep=""))
print(psFormula)


#=================================
#estimate propensity scores with logistic regression
library(survey)
surveyDesign <- svydesign(ids=~0,strata=~STRATA,weights=~FINALWGT, data=SSOCS.data)


psModel <- svyglm(psFormula, surveyDesign, family=quasibinomial)

#add propensity scores to dataset
SSOCS.data$pScores <- fitted(psModel)


#evaluate common support with plots
tiff("Chapter4_figure4-1.tif", res=600, compression = "lzw", height=6, width=15, units="in")
hist(SSOCS.data$pScores[SSOCS.data$treat==0], density = 10, angle = 45, main="Propensity Scores",
     xlab="Shaded = Untreated | Gray = Treated") 
hist(SSOCS.data$pScores[SSOCS.data$treat==1], col=gray(0.4,0.25), add=T) 
dev.off()

#========================================
library(lattice)
tiff("Chapter4_figure4-2.tif", res=600, compression = "lzw", height=6, width=15, units="in")
bwplot( pScores~treat, data = SSOCS.data, 
        ylab = "Propensity Scores", auto.key = TRUE)
dev.off()

#obtain descriptive statistics
summary(SSOCS.data$pScores[SSOCS.data$treat==0])
summary(SSOCS.data$pScores[SSOCS.data$treat==1])


# #=====================================================
#Estimate propensity scores with GBM
library(twang)
#change the treatment indicator from factor to numeric, which is required by GBM
set.seed(2016)
SSOCS.data$treat <- as.numeric(as.character(SSOCS.data$treat)) 
myGBM <- ps(psFormula, data = SSOCS.data, n.trees=10000, interaction.depth=4,
            shrinkage=0.01, stop.method=c("es.max"), estimand = "ATT", 
            verbose=TRUE, sampw = SSOCS.data$FINALWGT)

#obtain information about balance achieved and iteration numbers
summary(myGBM)


#obtain tiff image
tiff("Chapter4_GBM_convergence.tif", res=600, compression = "lzw", height=6, width=15, units="in")
plot(myGBM,type="b", color=F)
dev.off()

#extract estimated propensity scores from object
pScoresGBM <- myGBM$ps
names(pScoresGBM) = "pScoresGBM"
SSOCS.data$pScoresGBM <- pScoresGBM


#=========================================================

#cut propensity scores into five strata for estimating the ATT

SSOCS.data$subclass <- cut(x=SSOCS.data$pScores,
                              breaks=quantile(SSOCS.data$pScores, 
                              prob = seq(0, 1, 1/5)),include.lowest=T)
#rename the strata labels
levels(SSOCS.data$subclass) <- 1:length(levels(SSOCS.data$subclass))
#examine common support
xtabs(~treat+subclass,SSOCS.data) #table of counts per stratum

#save the data
#save(SSOCS.data, file="SSOCS_data_complete.Rdata", compress=T)


#=========================================================================
#evaluate covariate balance within  strata separately for categorical and continuous variables
library(MatchIt)

SSOCS.data2 <- SSOCS.data[,c("SCHID", "treat","percHarsh", "pScores","pScores",
                            "STRATA", "FINALWGT",covariateNames)]


balanceFormula <- paste(covariateNames, collapse="+")
balanceFormula <- formula(paste("treat~",balanceFormula,sep=""))
stratification <- matchit(balanceFormula,distance=SSOCS.data2$pScores, 
                          data = SSOCS.data2, 
                          method = "subclass", sub.by = "treat", subclass=5)
#the code above gives a warning that does not apply
data.stratification <- match.data(stratification)
data.stratification$treat <- factor(data.stratification$treat, levels=c(0,1),
                                   labels=c("Untreated","Treated"))
data.stratification$subclass <- factor(data.stratification$subclass)
xtabs(~treat+subclass, data.stratification)

#diagnose covariate balance
balance.stratification <- summary(stratification, standardize=T)

#obtain standardized mean differences for all strata
strataDifferences <- data.frame(balance.stratification$q.table[,3,])
summaryStrataDifferences <- summary(strataDifferences)
#write.csv(data.frame(summaryStrataDifferences), file="summary_strata_differences.csv")

#obtain the summary of balance after stratification combining all strata
summary(abs(balance.stratification$sum.subclass$"Std. Mean Diff."))
table(abs(balance.stratification$sum.subclass$"Std. Mean Diff.") > 0.1)

#=================================================

 library(twang)
#combine the sampling weight with the stratum weight
data.stratification$stratumWeight <- with(data.stratification, 
                                    FINALWGT*weights/mean(FINALWGT*weights))

balanceResults <- list() #empty storage

#loop through strata to perform balance evaluation
for (stratum in 1:5) { #loop through strata
 
    tempData <- data.stratification[data.stratification$subclass==stratum,]
    
    #balance evaluation for continuous covariates
    tableBalance <- bal.stat(
        data=tempData, estimand="ATT",  
        w.all=tempData$stratumWeight,get.ks=F, vars=covariateNames, 
        treat.var="treat", sampw=1, multinom=F)
    tableBalance <- tableBalance$results[,1:5]
  
#store results
balanceResults[[stratum]]<-tableBalance
    
} #close loop

#summarize covariate balance
balance.stratification2 <- lapply(balanceResults,summary)


#=================================================================================
#Estimating and Pooling Stratum-specific effects

#obtain means for treated and untreated groups by stratum
library(survey)

#define survey design

surveyDesign <- svydesign(ids=~0,strata=~STRATA,weights=~FINALWGT, data=data.stratification)
#add replication weights for 1000 bootstrapped samples to the design object
surveyDesignBoot <- as.svrepdesign(surveyDesign, type=c("bootstrap"),replicates=1000)

#obtain means and standard errors by stratum and treatment group combinations
(subclassMeans <- svyby(formula=~percHarsh, by=~treat+subclass,
                        design=surveyDesignBoot,FUN=svymean, covmat=TRUE))



#obtain the ATE pooling strata-specific ATEs
#the weights are for each group defined by combining stratum and treatment group
#the weights are (stratum size)/(total sample size)
(pooledEffects <- svycontrast(subclassMeans, list(
    ATE=c(-0.5,0.5,-0.18,0.18,-0.13,0.13,-0.10,0.10,-0.09,0.09),
    ATT=c(-0.2,0.2,-0.2,0.2,-0.2,0.2,-0.2,0.2,-0.2,0.2))))



#=================================
#regression adjustment
# regressionAdjuster = function(formula,design) {
#     model = svyglm(formula,design)
# (subclassMeansAdjusted <- svyby(percHarsh~ FR_CATMN, by=~treat+subclass,
#                         design=surveyDesignBoot,FUN=svyglm, covmat=TRUE,deff=F))




#===============================================================================
#Code to obtain marginal mean weights throughs stratification for the ATE
#obtain subclass for the ATE


#obtain the number of people per subclass by treatment combination
subclassbyTreat <- data.frame(xtabs(~treat+subclass,data.stratification))
names(subclassbyTreat)<-c("treat","subclass","N.zs")
print(subclassbyTreat)
#obtain the number of people per subclass
table.subclass<- data.frame(xtabs(~subclass,data.stratification))
names(table.subclass)<-c("subclass","N.s")
print(table.subclass)
#merge tables
(subclassbyTreat <- merge(subclassbyTreat,table.subclass))

#obtain estimates of proportion treated and untreated
(prop.treat <- svymean(~treat, surveyDesign))

#insert proportion treated into table
subclassbyTreat$pr.Z <- ifelse(subclassbyTreat$treat=="Treated",prop.treat[2],prop.treat[1])

#obtain the mmwts for ATE
subclassbyTreat$mmwsATE <- with(subclassbyTreat, N.s*pr.Z/N.zs)
print(subclassbyTreat)

data.stratification <- merge(data.stratification, subclassbyTreat[,c(1,2,6)]) #merge weights with data.stratification
xtabs(~mmwsATE+subclass, data.stratification)

#combine marginal mean weight and sampling weight, then normalize
data.stratification$mmwsATEFinal <- data.stratification$mmwsATE*data.stratification$FINALWGT 
data.stratification$mmwsATEFinal <- data.stratification$mmwsATEFinal/mean(data.stratification$mmwsATEFinal)


#================================================================
#code to obtain marginal mean weights through stratification for the ATT


#1. obtain the number of people per subclass for the treated
subclassTreat <- data.frame(table(data.stratification$subclass[data.stratification$treat=="Treated"]))
names(subclassTreat) <- c("subclass","N.1s")
print(subclassTreat)
#obtain the number of people per subclass for the untreated
subclassUntreat <-  data.frame(table(data.stratification$subclass[data.stratification$treat=="Untreated"]))
names(subclassUntreat) <- c("subclass","N.0s")
print(subclassUntreat)
#merge treated and untreated
(table.subclass <- merge(subclassTreat,subclassUntreat))

#merge table with the data
data.stratification <- merge(data.stratification,table.subclass)

#obtain proportions treated and untreated
(prop.treat <- svymean(~treat, surveyDesign))


#calculate the ATT MMWS
data.stratification$mmwsATT <- with(data.stratification,
        ifelse(treat=="Treated", 1, N.1s*prop.treat[1]/N.0s*prop.treat[2]))
xtabs(~mmwsATT+subclass, data.stratification)



#combine marginal mean weight and sampling weight, then normalize
data.stratification$mmwsATTFinal <- data.stratification$mmwsATT*data.stratification$FINALWGT 
data.stratification$mmwsATTFinal <- data.stratification$mmwsATTFinal/mean(data.stratification$mmwsATTFinal)

#==============================================
#Evaluate coariate balance for the ATE

library(twang)
#create a numeric treatment indicator, required by twang
data.stratification$treat2 <- ifelse(data.stratification$treat == "Treated", 1,0) 

#balance evaluation 
balanceTableMMWS <- bal.stat(
    data=data.stratification, 
    estimand="ATE",  w.all=data.stratification$mmwsATEFinal,get.ks=F, vars=covariateNames, 
    treat.var="treat2", sampw=data.stratification$FINALWGT, multinom=F)
balanceTableMMWS <- balanceTableMMWS$results[,1:5]
balanceTableMMWS$varRatio <- with(balanceTableMMWS, tx.sd^2/ct.sd^2) 

unbalanced.covariatesATE = list(
    std.eff.sz = rownames(balanceTableMMWS)[abs(balanceTableMMWS$std.eff.sz)>0.1],
    varRatio = rownames(balanceTableMMWS)[balanceTableMMWS$varRatio<0.8 |
                                              balanceTableMMWS$varRatio>1.2 ])


#examine balance statistics
summary(balanceTableMMWS)
summaryTableMMWS <- c()
for (var in 1:ncol(balanceTableMMWS)) {
    summaryTableMMWS = rbind(summaryTableMMWS,summary(balanceTableMMWS[,var])) }
rownames(summaryTableMMWS) = colnames(balanceTableMMWS)
#write.csv(summaryTableMMWS[,c(1,4,6)], file="summary_balance_MMWSATE.csv")




#=============================================
#evaluate covariate balance for the ATT
library(twang)
#create a numeric treatment indicator, required by twang
data.stratification$treat2 <- ifelse(data.stratification$treat == "Treated", 1,0) 

#balance evaluation 
balanceTableMMWSATT <- bal.stat(
    data=data.stratification, 
    estimand="ATT",  w.all=data.stratification$mmwsATTFinal,get.ks=F, vars=covariateNames, 
    treat.var="treat2", sampw=data.stratification$FINALWGT, multinom=F)
balanceTableMMWSATT <- balanceTableMMWSATT$results[,1:5]
balanceTableMMWSATT$varRatio <- with(balanceTableMMWSATT, tx.sd^2/ct.sd^2) 


#examine balance statistics
summary(balanceTableMMWSATT)
summaryTableMMWSATT = c()
for (var in 1:ncol(balanceTableMMWSATT)) {
    summaryTableMMWSATT = rbind(summaryTableMMWSATT,summary(balanceTableMMWSATT[,var])) }
rownames(summaryTableMMWSATT) = colnames(balanceTableMMWSATT)
write.csv(summaryTableMMWSATT[,c(1,4,6)], file="summary_balance_MMWSATT.csv")

unbalanced.covariatesATT <- list(
    std.eff.sz = rownames(balanceTableMMWSATT)[abs(balanceTableMMWSATT$std.eff.sz)>0.1],
    varRatio = rownames(balanceTableMMWSATT)[balanceTableMMWSATT$varRatio<0.8 |
                                              balanceTableMMWSATT$varRatio>1.2 ])
#=======================================================================


#estimate ATE using marginal mean weighting through stratification to adjust selection bias
surveyDesignATE <- svydesign(ids=~0,strata=~STRATA,weights=~mmwsATEFinal, data=data.stratification)
surveyDesignATE <- as.svrepdesign(surveyDesignATE, type=c("bootstrap"),replicates=1000)

#fit a linear regression model to proportions
modelATE <- svyglm(percHarsh~treat, surveyDesignATE, family=gaussian())
summary(modelATE)

#fit a logistic model  to proportions with quasibinomial family to allow for overdispersion
modelATE2 <- svyglm(percHarsh~treat, surveyDesignATE, family=quasibinomial())
summary(modelATE2)


#=======================================================================

#estimate ATT using marginal mean weighting through stratification to adjust selection bias
surveyDesignATT <- svydesign(ids=~0,strata=~STRATA,weights=~mmwsATTFinal, data=data.stratification)
surveyDesignATT <- as.svrepdesign(surveyDesignATT, type=c("bootstrap"),replicates=1000)

#fit a linear regression model to proportions
modelATT1 <- svyglm(percHarsh~treat, surveyDesignATT, family=gaussian())
summary(modelATT1)

#fit a logistic model  to proportions with quasibinomial family to allow for overdispersion
modelATT2 <- svyglm(percHarsh~treat, surveyDesignATT, family=quasibinomial())
summary(modelATT2)

#center continuous predictors on the mean of the treated
surveyDesignATT <- update(surveyDesignATT, C0540 = C0540 - mean(C0540[treat=="Treated"]),
                         C0568 = C0568 - mean(C0568[treat=="Treated"]),
                         C0544 = C0544 - mean(C0544[treat=="Treated"]),
                         C0558 = C0558 - mean(C0558[treat=="Treated"]))
modelATTDR <- svyglm(percHarsh~(treat+subclass+FR_SIZE+C0540+C0158+C0166+C0568+C0540+C0544+C0558)^2, 
                    surveyDesignATT, family=gaussian())
summary(modelATTDR)

#save(list=ls(), file="Chapter4_All_results.Rdata", compress=T)



